"""Optimizers."""

import autograd.numpy as np
import autograd
import scipy.optimize


class NewtonOptimizer(object):
    """An optimizer based on Newton's method."""

    def __init__(self, tol=1.48e-08, maxiter=50):
        """Generate an optimizer from an objective function."""
        self.tol = tol
        self.maxiter = maxiter

    def prepare(self, f):
        """Accept an objective function for optimization."""
        self.g = autograd.grad(f)
        self.h = autograd.hessian(f)

    def optimize(self, x0, target):
        """Calculate an optimum argument of an objective function."""
        x = x0
        for _ in range(self.maxiter):
            delta = np.linalg.solve(self.h(x, target), -self.g(x, target))
            x = x + delta
            if np.linalg.norm(delta) < self.tol:
                break
        return x


class SteepestDescentOptimizer(object):
    """An optimizer based on steepest descent method."""

    def __init__(self, tol=1.48e-08, maxiter=50, alpha=1):
        """Generate an optimizer from an objective function."""
        self.tol = tol
        self.maxiter = maxiter
        self.alpha = alpha

    def prepare(self, f):
        """Accept an objective function for optimization."""
        self.g = autograd.grad(f)

    def optimize(self, x0, target):
        """Calculate an optimum argument of an objective function."""
        x = x0
        for _ in range(self.maxiter):
            delta = self.alpha * self.g(x, target)
            x = x - delta
            if np.linalg.norm(delta) < self.tol:
                break
        return x


class ConjugateGradientOptimizer(object):
    """An optimizer based on conjugate gradient method."""

    def __init__(self, tol=1.48e-08, maxiter=50):
        """Generate an optimizer from an objective function."""
        self.tol = tol
        self.maxiter = maxiter

    def prepare(self, f):
        """Accept an objective function for optimization."""
        self.g = autograd.grad(f)
        self.h = autograd.hessian(f)

    def optimize(self, x0, target):
        """Calculate an optimum argument of an objective function."""
        x = x0
        for i in range(self.maxiter):
            g = self.g(x, target)
            h = self.h(x, target)
            if i == 0:
                alpha = 0
                m = g
            else:
                alpha = - np.dot(m, np.dot(h, g)) / np.dot(m, np.dot(h, m))
                m = g + np.dot(alpha, m)
            t = - np.dot(m, g) / np.dot(m, np.dot(h, m))
            delta = np.dot(t, m)
            x = x + delta
            if np.linalg.norm(delta) < self.tol:
                break
        return x


class ScipyOptimizer(object):
    """An optimizer based on scipy.optimize.minimize."""

    def __init__(self, **optimizer_opt):
        """Generate an optimizer from an objective function."""
        for k, v in [  # default values
                ('method', 'BFGS'),
                ('tol', 1.48e-08),
                ('options', {'maxiter': 50})]:
            if k not in optimizer_opt:
                optimizer_opt[k] = v
        self.optimizer_opt = optimizer_opt

    def prepare(self, f):
        """Accept an objective function for optimization."""
        self.f = f

    def optimize(self, angles0, target):
        """Calculate an optimum argument of an objective function."""
        def new_objective(angles):
            return self.f(angles, target)

        return scipy.optimize.minimize(
            new_objective,
            angles0,
            **self.optimizer_opt).x


class ScipySmoothOptimizer(ScipyOptimizer):
    """A smooth optimizer based on scipy.optimize.minimize."""

    def __init__(self, smooth_factor=.1, **optimizer_opt):
        """Generate an optimizer from an objective function."""
        self.smooth_factor = smooth_factor
        for k, v in [  # default values
                ('method', 'L-BFGS-B'),
                ('bounds', None)]:
            if k not in optimizer_opt:
                optimizer_opt[k] = v
        super(ScipySmoothOptimizer, self).__init__(**optimizer_opt)

    def optimize(self, angles0, target):
        """Calculate an optimum argument of an objective function."""
        def new_objective(angles):
            a = angles - angles0
            if isinstance(self.smooth_factor, (np.ndarray, list)):
                if len(a) == len(self.smooth_factor):
                    return (self.f(angles, target) +
                            np.sum(self.smooth_factor * np.power(a, 2)))
                else:
                    raise ValueError('len(smooth_factor) != number of joints')
            else:
                return (self.f(angles, target) +
                        self.smooth_factor * np.sum(np.power(a, 2)))

        return scipy.optimize.minimize(
            new_objective,
            angles0,
            **self.optimizer_opt).x



"""Solvers."""

from functools import reduce
import sys

import autograd.numpy as np

from .component import Joint


class FKSolver(object):
    """A forward kinematics solver."""

    def __init__(self, components):
        """Generate a FK solver from link and joint instances."""
        joint_indexes = [
            i for i, c in enumerate(components) if isinstance(c, Joint)
        ]

        def matrices(angles):
            joints = dict(zip(joint_indexes, angles))
            a = [joints.get(i, None) for i in range(len(components))]
            return [c.matrix(a[i]) for i, c in enumerate(components)]

        self._matrices = matrices

    def solve(self, angles):
        """Calculate a position of the end-effector and return it."""
        return reduce(
            lambda a, m: np.dot(m, a),
            reversed(self._matrices(angles)),
            np.array([0., 0., 0., 1.])
        )[:3]


class IKSolver(object):
    """An inverse kinematics solver."""

    def __init__(self, fk_solver, optimizer):
        """Generate an IK solver from a FK solver instance."""
        def distance_squared(angles, target):
            x = target - fk_solver.solve(angles)
            return np.sum(np.power(x, 2))

        optimizer.prepare(distance_squared)
        self.optimizer = optimizer

    def solve(self, angles0, target):
        """Calculate joint angles and returns it."""
        return self.optimizer.optimize(np.array(angles0), target)


class CCDFKSolver(object):

    def __init__(self, components):
        joint_indexes = [
            i for i, c in enumerate(components) if isinstance(c, Joint)
        ]

        def matrices(angles):
            joints = dict(zip(joint_indexes, angles))
            a = [joints.get(i, None) for i in range(len(components))]
            return [c.matrix(a[i]) for i, c in enumerate(components)]

        self._matrices = matrices
        self.components = components
        self.joint_indexes = joint_indexes

    def solve(self, angles, p=None, index=None):
        if p is None:
            p = [0., 0., 0., 1.]
        if index is None:
            index = len(self.components) - 1
        return reduce(
            lambda a, m: np.dot(m, a),
            reversed(self._matrices(angles)[:index + 1]),
            np.array(p)
        )[:3]


class CCDIKSolver(object):

    def __init__(self, fk_solver, tol=1.48e-08, maxiter=50):
        self._fk_solver = fk_solver
        self.tol = tol
        self.maxiter = maxiter

    def solve(self, angles0, target):
        joint_indexes = list(reversed(self._fk_solver.joint_indexes))
        angles = angles0[:]
        pmap = {
            'x': [1., 0., 0., 0.],
            'y': [0., 1., 0., 0.],
            'z': [0., 0., 1., 0.]}
        prev_dist = sys.float_info.max
        for _ in range(self.maxiter):
            for i, idx in enumerate(joint_indexes):
                axis = self._fk_solver.solve(
                    angles,
                    p=pmap[self._fk_solver.components[idx].axis],
                    index=idx)
                pj = self._fk_solver.solve(angles, index=idx)
                ee = self._fk_solver.solve(angles)
                eorp = self.p_on_rot_plane(ee, pj, axis)
                torp = self.p_on_rot_plane(target, pj, axis)
                ve = (eorp - pj) / np.linalg.norm(eorp - pj)
                vt = (torp - pj) / np.linalg.norm(torp - pj)
                a = np.arccos(np.dot(vt, ve))
                sign = 1 if np.dot(axis, np.cross(ve, vt)) > 0 else -1
                angles[-(i + 1)] += (a * sign)

            ee = self._fk_solver.solve(angles)
            dist = np.linalg.norm(target - ee)
            delta = prev_dist - dist
            if delta < self.tol:
                break
            prev_dist = dist

        return angles

    def p_on_rot_plane(self, p, joint_pos, joint_axis):
        ua = joint_axis / np.linalg.norm(joint_axis)
        return p - (np.dot(p - joint_pos, ua) * ua)



"""Components for an actuator."""

import autograd.numpy as np


class Link(object):
    """Represents a link."""

    def __init__(self, coord):
        """Create a link from a specified coordinate."""
        self.coord = coord

    def matrix(self, _):
        """Return translation matrix in homogeneous coordinates."""
        x, y, z = self.coord
        return np.array([
            [1., 0., 0., x],
            [0., 1., 0., y],
            [0., 0., 1., z],
            [0., 0., 0., 1.]
        ])


class Joint(object):
    """Represents a revolute joint."""

    def __init__(self, axis):
        """Create a revolute joint from a specified axis."""
        self.axis = axis

    def matrix(self, angle):
        """Return rotation matrix in homogeneous coordinates."""
        _rot_mat = {
            'x': self._x_rot,
            'y': self._y_rot,
            'z': self._z_rot
        }
        return _rot_mat[self.axis](angle)

    def _x_rot(self, angle):
        return np.array([
            [1., 0., 0., 0.],
            [0., np.cos(angle), -np.sin(angle), 0.],
            [0., np.sin(angle), np.cos(angle), 0.],
            [0., 0., 0., 1.]
        ])

    def _y_rot(self, angle):
        return np.array([
            [np.cos(angle), 0., np.sin(angle), 0.],
            [0., 1., 0., 0.],
            [-np.sin(angle), 0., np.cos(angle), 0.],
            [0., 0., 0., 1.]
        ])

    def _z_rot(self, angle):
        return np.array([
            [np.cos(angle), -np.sin(angle), 0., 0.],
            [np.sin(angle), np.cos(angle), 0., 0.],
            [0., 0., 1., 0.],
            [0., 0., 0., 1.]
        ])



import numpy as np

try:
    import open3d as o3d  # the extra feature
except ImportError:
    pass


def translate(p):
    x, y, z = p
    return np.array([
            [1., 0., 0., x],
            [0., 1., 0., y],
            [0., 0., 1., z],
            [0., 0., 0., 1.]
        ])


def rotate(axis, angle):
    x, y, z = axis
    return np.array([  # Rodrigues
            [
                np.cos(angle) + (x**2 * (1 - np.cos(angle))),
                (x * y * (1 - np.cos(angle))) - (z * np.sin(angle)),
                (x * z * (1 - np.cos(angle))) + (y * np.sin(angle)),
                0.
            ], [
                (y * x * (1 - np.cos(angle))) + (z * np.sin(angle)),
                np.cos(angle) + (y**2 * (1 - np.cos(angle))),
                (y * z * (1 - np.cos(angle))) - (x * np.sin(angle)),
                0.
            ], [
                (z * x * (1 - np.cos(angle))) - (y * np.sin(angle)),
                (z * y * (1 - np.cos(angle))) + (x * np.sin(angle)),
                np.cos(angle) + (z**2 * (1 - np.cos(angle))),
                0.
            ], [0., 0., 0., 1.]
        ])


def create_sphere(p, r=.1, color=None):
    if color is None:
        color = [.8, .8, .8]
    geo = o3d.geometry.TriangleMesh.create_sphere(radius=r)
    geo.compute_vertex_normals()
    geo.paint_uniform_color(color)
    geo.transform(translate(p))
    return geo


class GeoComponent:

    child = None

    def tip(self, link_color=None):
        return create_sphere(
            [0., 0., 0.],
            color=[.8, .8, .8] if link_color is None else link_color)

    def geo(self, mat=None, link_color=None):
        geo = self.base_geo(link_color)
        if mat is not None:
            geo.transform(mat)
            mat = mat @ self.mat()
        else:
            mat = self.mat()
        if self.child is None:
            return [geo] + [self.tip(link_color).transform(mat)]
        else:
            return [geo] + self.child.geo(mat, link_color)


class Link(GeoComponent):

    def __init__(self, c):
        self.c = c

    def base_geo(self, link_color=None):
        norm = np.linalg.norm(self.c.coord)
        base = np.array([0., 0., norm])
        cross = np.cross(base, self.c.coord)
        axis = cross / np.linalg.norm(cross)
        angle = np.arccos(np.dot(base, self.c.coord) / (norm ** 2))
        geo = o3d.geometry.TriangleMesh.create_cylinder(
            radius=.05, height=norm)
        geo.compute_vertex_normals()
        geo.paint_uniform_color(
            [.8, .8, .8] if link_color is None else link_color)
        geo.transform(rotate(axis, angle) @ translate(base / 2))
        return geo

    def mat(self):
        return self.c.matrix(None)


class Joint(GeoComponent):

    def __init__(self, c):
        self.c = c
        self.angle = 0.

    def base_geo(self, _=None):
        geo = o3d.geometry.TriangleMesh.create_cylinder(radius=.1, height=.2)
        geo.compute_vertex_normals()
        geo.paint_uniform_color([.2, .2, .9])
        rx = {
            'x': [0., 1., 0.],
            'y': [1., 0., 0.],
            'z': [0., 0., 1.],
        }
        geo.transform(rotate(rx[self.c.axis], np.pi / 2))
        return geo

    def mat(self):
        return self.c.matrix(self.angle)


def visualize(actuator, target=None):
    root = None
    p = None
    joints = []
    for c in actuator.components:
        if hasattr(c, 'axis'):
            gc = Joint(c)
            joints.append(gc)
        else:
            gc = Link(c)

        if root is None:
            root = gc
            p = gc
        else:
            p.child = gc
            p = gc

    for j, a in zip(joints, actuator.angles):
        j.angle = a

    if target:
        geos = root.geo(link_color=[.5, .5, .5])
        actuator.ee = target
        for j, a in zip(joints, actuator.angles):
            j.angle = a
        geos += root.geo()
        geos += [create_sphere(target, r=.12, color=[.8, .2, .2])]
    else:
        geos = root.geo()

    o3d.visualization.draw_geometries(
        geos, window_name='tinyik vizualizer', width=640, height=480)



"""Core features."""

from numbers import Number

import autograd.numpy as np

from .component import Link, Joint
from .solver import FKSolver, IKSolver
from .optimizer import ScipyOptimizer


class Actuator(object):
    """Represents an actuator as a set of links and revolute joints."""

    def __init__(self, tokens, optimizer=ScipyOptimizer()):
        """Create an actuator from specified link lengths and joint axes."""
        components = []
        for t in tokens:
            if isinstance(t, Number):
                components.append(Link([t, 0., 0.]))
            elif isinstance(t, list) or isinstance(t, np.ndarray):
                components.append(Link(t))
            elif isinstance(t, str) and t in {'x', 'y', 'z'}:
                components.append(Joint(t))
            else:
                raise ValueError(
                    'the arguments need to be '
                    'link length or joint axis: {}'.format(t)
                )

        self.fk = FKSolver(components)
        self.ik = IKSolver(self.fk, optimizer)

        self.angles = [0.] * len(
            [c for c in components if isinstance(c, Joint)]
        )
        self.components = components

    @property
    def angles(self):
        """The joint angles."""
        return self._angles

    @angles.setter
    def angles(self, angles):
        self._angles = np.array(angles)

    @property
    def ee(self):
        """The end-effector position."""
        return self.fk.solve(self.angles)

    @ee.setter
    def ee(self, position):
        self.angles = self.ik.solve(self.angles, position)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tinyik']

package_data = \
{'': ['*']}

install_requires = \
['autograd', 'scipy']

extras_require = \
{'viz': ['open3d>=0.9.0,<0.10.0']}

setup_kwargs = {
    'name': 'tinyik',
    'version': '2.1.0',
    'description': 'A tiny inverse kinematics solver',
    'long_description': "tinyik\n======\n\ntinyik is a simple and naive inverse kinematics solver.\n\nIt defines the actuator as a set of links and revolute joints from an origin. Here is the example of a robot arm that consists of two joints that rotate around z-axis and two links of 1.0 length along x-axis:\n\n.. code-block:: python\n\n    >>> import tinyik\n    >>> arm = tinyik.Actuator(['z', [1., 0., 0.], 'z', [1., 0., 0.]])\n\nSince the joint angles are zero by default, the end-effector position is at (2.0, 0, 0):\n\n.. code-block:: python\n\n    >>> arm.angles\n    array([ 0.,  0.])\n    >>> arm.ee\n    array([ 2.,  0.,  0.])\n\nSets the joint angles to 30 and 60 degrees to calculate a new position of the end-effector:\n\n.. code-block:: python\n\n    >>> import numpy as np\n    >>> arm.angles = [np.pi / 6, np.pi / 3]  # or np.deg2rad([30, 60])\n    >>> arm.ee\n    array([ 0.8660254,  1.5      ,  0.       ])\n\nSets a position of the end-effector to calculate the joint angles:\n\n.. code-block:: python\n\n    >>> arm.ee = [2 / np.sqrt(2), 2 / np.sqrt(2), 0.]\n    >>> arm.angles\n    array([  7.85398147e-01,   3.23715739e-08])\n    >>> np.round(np.rad2deg(arm.angles))\n    array([ 45.,   0.])\n\nOptionally, it has the visualization feature. Passes the actuator to it to visualize its structure:\n\n.. code-block:: python\n\n    >>> leg = tinyik.Actuator([[.3, .0, .0], 'z', [.3, .0, .0], 'x', [.0, -.5, .0], 'x', [.0, -.5, .0]])\n    >>> leg.angles = np.deg2rad([30, 45, -90])\n    >>> tinyik.visualize(leg)\n\n.. image:: https://raw.githubusercontent.com/lanius/tinyik/master/assets/viz_structure.png\n\nPasses with the target position, can compare before and after the IK. The gray links are before IK and the white links are after it. The red sphere is the target position:\n\n.. code-block:: python\n\n    >>> tinyik.visualize(leg, target=[.8, .0, .8])\n\n.. image:: https://raw.githubusercontent.com/lanius/tinyik/master/assets/viz_ik.png\n\nInstallation\n------------\n\n.. code-block:: console\n\n    $ pip install tinyik\n\nWith the visualization feature:\n\n.. code-block:: console\n\n    $ pip install tinyik[viz]\n",
    'author': 'lanius',
    'author_email': 'lanius@nirvake.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lanius/tinyik',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



"""A simple and naive inverse kinematics solver."""

from .core import Actuator
from .component import Link, Joint
from .solver import FKSolver, IKSolver, CCDFKSolver, CCDIKSolver
from .optimizer import (
    NewtonOptimizer, SteepestDescentOptimizer, ConjugateGradientOptimizer,
    ScipyOptimizer, ScipySmoothOptimizer
)
from .visualizer import visualize


__all__ = (
    'Actuator',
    'Link', 'Joint',
    'FKSolver', 'IKSolver', 'CCDFKSolver', 'CCDIKSolver',
    'NewtonOptimizer', 'SteepestDescentOptimizer',
    'ConjugateGradientOptimizer',
    'ScipyOptimizer', 'ScipySmoothOptimizer',
    'visualize'
)
